#include "stdafx.h"
#include "IRProviderIR.hpp"
#include "utils.hpp"
#include <fstream>
#include <harkio2/harkio.hpp>

namespace
{

// File1がmic_indexだけをマイクインデックスとして持つかどうかを示す述語.
struct HavingOneMicIndex
{
    HavingOneMicIndex(unsigned int mic_index) : mic_index_(mic_index) { }
    bool operator()(const harkio2::harkconfig::File1& file) const
    {
        return file.IsMicIndicesInitialized() &&
            file.GetMicIndices().size() == 1 &&
            file.GetMicIndices().front() == mic_index_;
    }
private:
    const unsigned int mic_index_;
};

}

namespace harktool4
{

class IRProviderIR::Impl
{
public:
    Impl(const std::string& ir_list_file);
    std::vector<float> GetIR(size_t src_index, size_t mic_index);
    size_t GetNumSrc();
    size_t GetNumMic();
    std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> GetSources();

private:
    boost::shared_ptr<harkio2::harkconfig::ImpulseResponseListFile> ir_;
};

IRProviderIR::Impl::Impl(const std::string& ir_list_file)
{
    // ir_list_fileというファイル名のファイルを、HARK-configファイルだと思って開く
    harkio2::harkconfig::HarkConfigFileXMLReader reader(ir_list_file);

    // ファイル内容を読み取る。ファイル内容が適切なフォーマットでないなら例外が発生。
    // 例外はあえてキャッチせず、Implコンストラクタ呼び出し側に対処を任せる
    auto hark_config_file = reader.Read();

    Requires<std::runtime_error>(hark_config_file, "ImpulseResponseListFile", { harkio2::Version(2, 0) });
    /*
    if (!hark_config_file || hark_config_file->GetId() != "ImpulseResponseListFile")
    {
        throw std::runtime_error((boost::format("%s is not ImpulseResponseListFile")
                    % ir_list_file).str());
    }
    */

    ir_ = boost::dynamic_pointer_cast<
        harkio2::harkconfig::ImpulseResponseListFile>(hark_config_file);

    // hark_config_file->GetId()をチェックしているので、キャストに失敗することはあり得ないはず。
    assert(ir_);
}

std::vector<float> IRProviderIR::Impl::GetIR(size_t src_index, size_t mic_index)
{
    if (ir_->GetVersion() == harkio2::Version(2, 0))
    {
        auto ir = boost::dynamic_pointer_cast<harkio2::harkconfig::ImpulseResponseListFile2_0>(ir_);
        const auto& sources = ir->GetSources();
        const auto& source = sources.at(src_index); // 自動で範囲チェックをするためにat関数を使う
        const auto& files = source.GetFiles();

        // mic_indexを持つファイルをさがす
        auto file_having_mic = std::find_if(
                files.begin(), files.end(), HavingOneMicIndex(mic_index));

        // ファイルを開き、インパルス応答を取得する
        if (file_having_mic != files.end())
        {
            const auto& file_name = file_having_mic->GetName();
            std::ifstream ifs(file_name.c_str(), std::ios::binary);
            return ReadFloatArray(ifs);
        }
    }

    throw std::runtime_error("not implemented");
}

size_t IRProviderIR::Impl::GetNumSrc()
{
    if (ir_->GetVersion() == harkio2::Version(2, 0))
    {
        auto ir = boost::dynamic_pointer_cast<harkio2::harkconfig::ImpulseResponseListFile2_0>(ir_);
        return ir->GetSources().size();
    }

    throw std::runtime_error("not implemented");
}

size_t IRProviderIR::Impl::GetNumMic()
{
    if (ir_->GetVersion() == harkio2::Version(2, 0))
    {
        auto ir = boost::dynamic_pointer_cast<harkio2::harkconfig::ImpulseResponseListFile2_0>(ir_);
        if (ir->IsNumMicInitialized())
        {
            return ir->GetNumMic();
        }
        else if (ir->GetSources().size() > 0)
        {
            return ir->GetSources().front().GetFiles().size();
        }
    }

    throw std::runtime_error("not implemented");
}

std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> IRProviderIR::Impl::GetSources()
{
    if (ir_->GetVersion() == harkio2::Version(2, 0))
    {
        auto ir = boost::dynamic_pointer_cast<harkio2::harkconfig::ImpulseResponseListFile2_0>(ir_);
        const auto& sources = ir->GetSources();
        std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> result;

        result.reserve(sources.size());
        for (auto it = sources.begin(); it != sources.end(); ++it)
        {
            result.push_back(boost::make_shared<harkio2::harkconfig::Source2>(*it));
        }
        return result;
    }

    throw std::runtime_error("not implemented");
}

IRProviderIR::IRProviderIR(const std::string& ir_list_file)
    : impl_(new Impl(ir_list_file))
{ }

std::vector<float> IRProviderIR::GetIR(size_t src_index, size_t mic_index)
{
    return impl_->GetIR(src_index, mic_index);
}

size_t IRProviderIR::GetNumSrc()
{
    return impl_->GetNumSrc();
}

size_t IRProviderIR::GetNumMic()
{
    return impl_->GetNumMic();
}

std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> IRProviderIR::GetSources()
{
    return impl_->GetSources();
}

}
