#include "stdafx.h"
#include "IRProviderTSP.hpp"
#include "utils.hpp"
#include <fstream>
#include <algorithm>
#include <stdexcept>
#include <boost/format.hpp>
#include <boost/filesystem.hpp>
#include <harkio2/harkio.hpp>

namespace fs = boost::filesystem;

namespace
{

// File1がmic_indexをマイクインデックスとして持つかどうかを示す述語.
struct HavingMicIndex
{
    HavingMicIndex(unsigned int mic_index) : mic_index_(mic_index) { }
    bool operator()(const harkio2::harkconfig::File1& file) const
    {
        if (file.IsMicIndicesInitialized())
        {
            const auto& indices = file.GetMicIndices();
            return std::find(indices.begin(), indices.end(), mic_index_) != indices.end();
        }
        return false;
    }
private:
    const unsigned int mic_index_;
};

int CalcSysImp(const harkio2::TypedFilename& original_impulse_file,
        unsigned int signal_max, unsigned int synchronous_average, unsigned int tsp_offset,
        const std::vector<unsigned int>& mic_indices, const fs::path& tsp_file, const fs::path& impulse_output_prefix)
{
	int status ;
    const char* file_type = 0;
    switch (original_impulse_file.GetType())
    {
    case harkio2::FileType::WAVE: file_type = "wav"; break;
    case harkio2::FileType::FLOAT: file_type = "flt"; break;
    default: ;
    }

    auto channel_list = harktool4::Join(mic_indices.begin(), mic_indices.end(), ",");


    auto command = (boost::format("%s --tsporig%s \"%s\" --sigmax \"%s\" --syncaddfreq \"%s\" --syncaddoffset \"%s\" --outchannellist \"%s\" \"%s\" \"%s\"")
                        % "calcSysImp"
                        % file_type
                        % original_impulse_file.GetName()
                        % signal_max
                        % synchronous_average
                        % tsp_offset
                        % channel_list
                        % tsp_file.native()
                        % impulse_output_prefix.native()
                        ).str();
    
    std::cout << "CalcSysImp: " << command << std::endl;
    
    status = system(command.c_str());

   /* 終了ステータスのチェック */
	if (WIFEXITED(status)) 
	{
		switch( WEXITSTATUS(status) )
		{
		case 127:
			printf("CalcSysImp is not found : status(%d) WEXITSTATUS(%d)\n", status, WEXITSTATUS(status) );
			break;
		default:		
			printf("status(%d) %s\n", status, strerror(WEXITSTATUS(status)));
			break;
		}
	}
	
	if (WIFSIGNALED(status)) 
	{
		printf("CalcSysImp error: signal no(%d)\n", WTERMSIG(status));
	}
	
    printf("\n");
    printf("\n");
    
      
	return status;

	//org    return system(command.c_str());
}

}

namespace harktool4
{

class IRProviderTSP::Impl
{
public:
    Impl(const std::string& tsp_list_file);
    Impl(const Impl&) = delete;
    Impl& operator=(const Impl&) = delete;

    ~Impl();

    std::vector<float> GetIR(size_t src_index, size_t mic_index);
    size_t GetNumSrc();
    size_t GetNumMic();
    std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> GetSources();

private:
    fs::path CreateFloatFilePrefix(const std::string& tsp_file_name);
    bool CreateFloatFiles(const harkio2::TypedFilename& original_impulse_file,
        unsigned int signal_max, unsigned int synchronous_average, unsigned int tsp_offset,
        const std::vector<unsigned int>& mic_indices, const std::string& tsp_file_name);

private:
    boost::shared_ptr<harkio2::harkconfig::TSPListFile> tsp_;
    fs::path temp_dir_path_;
};

IRProviderTSP::Impl::Impl(const std::string& tsp_list_file)
    : tsp_(), temp_dir_path_()
{
    // tsp_list_fileというファイル名のファイルを、HARK-configファイルだと思って開く
    harkio2::harkconfig::HarkConfigFileXMLReader reader(tsp_list_file);

    // ファイル内容を読み取る。ファイル内容が適切なフォーマットでないなら例外が発生。
    // 例外はあえてキャッチせず、Implコンストラクタ呼び出し側に対処を任せる
    auto hark_config_file = reader.Read();

    Requires<std::runtime_error>(hark_config_file, "TSPListFile", { harkio2::Version(2, 0) });
    /*
    if (!hark_config_file || hark_config_file->GetId() != "TSPListFile")
    {
        throw std::runtime_error((boost::format("%s is not TSPListFile")
                    % tsp_list_file).str());
    }
    */

    tsp_ = boost::dynamic_pointer_cast<
        harkio2::harkconfig::TSPListFile>(hark_config_file);

    // hark_config_file->GetId()をチェックしているので、キャストに失敗することはあり得ないはず。
    assert(tsp_);
}

std::vector<float> IRProviderTSP::Impl::GetIR(size_t src_index, size_t mic_index)
{
    if (tsp_->GetVersion() == harkio2::Version(2, 0))
    {
        auto tsp = boost::dynamic_pointer_cast<harkio2::harkconfig::TSPListFile2_0>(tsp_);
        const auto& sources = tsp->GetSources();
        const auto& source = sources.at(src_index); // 自動で範囲チェックをするためにat関数を使う
        const auto& files = source.GetFiles();

        if (temp_dir_path_.empty())
        {
            // FLTファイルを生成するための一時ディレクトリを作成
            temp_dir_path_ = CreateTemporaryDirectory();
            std::cerr << "Created a temporary directore: " << temp_dir_path_.native() << std::endl;
        }

        // mic_indexを持つファイルをさがす
        auto file_having_mic = std::find_if(
                files.begin(), files.end(), HavingMicIndex(mic_index));

        if (file_having_mic != files.end())
        {
            auto tsp_file_name = file_having_mic->GetName();
            // TSP -> IR変換をする
            // （既に file_having_mic->GetName() が表すTSPファイルについてfloatファイルを計算済みなら、何もしない）
            CreateFloatFiles(tsp->GetOriginalImpulseFile(), tsp->GetSignalMax(),
                    tsp->GetSynchronousAverage(), tsp->GetTSPOffset(),
                    file_having_mic->GetMicIndices(), tsp_file_name);

            // 変換されたfloatファイルを開く
            auto flt_file_name_prefix = CreateFloatFilePrefix(tsp_file_name);
            auto flt_file_name = (boost::format("%s_ch%d.flt") % flt_file_name_prefix.native() % (mic_index + 1)).str();
            std::cerr << "flt_file_name = " << flt_file_name << std::endl;
            std::ifstream ifs(
                    flt_file_name,
                    std::ios::binary);
            return ReadFloatArray(ifs);
        }

        /*
        // ファイルを開き、インパルス応答を取得する
        if (file_having_mic != files.end())
        {
            const auto& file_name = file_having_mic->GetName();
            std::ifstream ifs(file_name.c_str(), std::ios::binary);
            return ReadFloatArray(ifs);
        }
        */
    }

    throw std::runtime_error("not implemented");
}

size_t IRProviderTSP::Impl::GetNumSrc()
{
    if (tsp_->GetVersion() == harkio2::Version(2, 0))
    {
        auto tsp = boost::dynamic_pointer_cast<harkio2::harkconfig::TSPListFile2_0>(tsp_);
        return tsp->GetSources().size();
    }

    throw std::runtime_error("not implemented");
}

size_t IRProviderTSP::Impl::GetNumMic()
{
    if (tsp_->GetVersion() == harkio2::Version(2, 0))
    {
        auto tsp = boost::dynamic_pointer_cast<harkio2::harkconfig::TSPListFile2_0>(tsp_);
        if (tsp->IsNumMicInitialized())
        {
            return tsp->GetNumMic();
        }
        else if (tsp->GetSources().size() > 0)
        {
            const auto& files = tsp->GetSources().front().GetFiles();
            size_t nummic = 0;
            for (auto it = files.begin(); it != files.end(); ++it)
            {
                nummic += it->GetMicIndices().size();
            }
            return nummic;
        }
    }

    throw std::runtime_error("not implemented");
}

std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> IRProviderTSP::Impl::GetSources()
{
    if (tsp_->GetVersion() == harkio2::Version(2, 0))
    {
        auto tsp = boost::dynamic_pointer_cast<harkio2::harkconfig::TSPListFile2_0>(tsp_);
        const auto& sources = tsp->GetSources();
        std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> result;

        result.reserve(sources.size());
        for (auto it = sources.begin(); it != sources.end(); ++it)
        {
            result.push_back(boost::make_shared<harkio2::harkconfig::Source2>(*it));
        }
        return result;
    }

    throw std::runtime_error("not implemented");
}

fs::path IRProviderTSP::Impl::CreateFloatFilePrefix(const std::string& tsp_file_name)
{
    return temp_dir_path_ / tsp_file_name / "float";
}

bool IRProviderTSP::Impl::CreateFloatFiles(const harkio2::TypedFilename& original_impulse_file,
    unsigned int signal_max, unsigned int synchronous_average, unsigned int tsp_offset,
    const std::vector<unsigned int>& mic_indices, const std::string& tsp_file_name)
{
    /*
     * まだsound_file_nameからFLTファイルを生成していないならば生成する
     *
     * /tmp/sound_file_name/float_ch0.flt
     * というイメージ
     */

    // floatファイルを生成するディレクトリ
    fs::path dir_path = temp_dir_path_ / tsp_file_name;

    // dir_pathが見つからないなら、まだfloatファイルを作っていないという証拠
    if (!fs::exists(dir_path))
    {
        boost::system::error_code ec;
        std::cerr << "creating directory: " << dir_path.native() << std::endl;
        if (fs::create_directories(dir_path, ec))
        {
            // dir_pathディレクトリ作成成功
            CalcSysImp(original_impulse_file, signal_max,
                    synchronous_average, tsp_offset,
                    mic_indices, tsp_file_name,
                    CreateFloatFilePrefix(tsp_file_name));
            return true;
        }
        else
        {
            throw std::runtime_error(ec.message());
        }
    }

    return false;
}

IRProviderTSP::Impl::~Impl()
{
    //fs::remove_all(temp_dir_path_);
}

IRProviderTSP::IRProviderTSP(const std::string& tsp_list_file)
    : impl_(new Impl(tsp_list_file))
{ }

std::vector<float> IRProviderTSP::GetIR(size_t src_index, size_t mic_index)
{
    return impl_->GetIR(src_index, mic_index);
}

size_t IRProviderTSP::GetNumSrc()
{
    return impl_->GetNumSrc();
}

size_t IRProviderTSP::GetNumMic()
{
    return impl_->GetNumMic();
}

std::vector<boost::shared_ptr<harkio2::harkconfig::Source>> IRProviderTSP::GetSources()
{
    return impl_->GetSources();
}

}
